#!/usr/bin/perl

# "select.pl", a script for use with the Forum News Gateway
# copyright 1995 by Swarthmore College

# Given a list on standard input of account information as generated
# by "Scripts/accounts.pl", select some of the accounts according to whether they
# have specified properties, writing the selected information to standard out.

# The properties may be:
#	-cb <date>		whether the account was created before the given date
#	-ca <date>		whether the account was created after the given date
#	-ub <date>		whether the account was last used before the given date
#	-ua <date>		whether the account was last used after the given date
#	+<n> <regexp>	whether the item in column <n> matches the regular expression
#	-<n> <regexp>	whether the item in column <n> does not match the regular expression

# If you give more than one test on the command line, accounts are passed
# through only if they meet all tests.

# You can use this to, for example, pick out accounts which have been idle for
# three months and can be deleted, or which have unlikely-looking e-mail addresses.

# <date> can be in any of these forms:
#	10 Feb 1995				10 Feb				10
#	Feb 10 1995				Feb 10
#	2.10.95					2.10
# If you leave out the year, it assumes this year. If you leave out
# the month, it assumes this month. You can use just about any sort
# of punctuation. You can give the year either as "95" or as "1995".

# <regexp> is a perl regular expression

# The creation date (used by -cb and -ca) is assumed to be in column 5.
# The last-used date (used by -ub and -ua) is assumed to be in column 6.
# These are the columns where the information is put by "accounts.pl",
# but if you use "columns.pl" you may move them around.

# Accounts which have missing creation or use dates are assumed to
# be "before" any date that you give.

require "timelocal.pl";			# from the perl library

unshift (@INC,'Scripts');		# get "require" files from Scripts directory
require "header.ph";			# for $Header1 and $Header2

unshift (@INC,".");				# get "require" files from parent directory
require "time.pl";				# for %Months

# --

$errorMessage = "";

if (!&ProcessArgs()) {
	if ($errorMessage ne "") {
		print STDERR "$0: $errorMessage\n";
	} else {
		print STDERR "usage: perl $0 [<n1> [<n2> ...]]\n",
					 "    select specified columns from tab-delimited input\n",
					 "    e.g. 'perl $0 1 3 1' retains columns 1, 3 and 1 in that order\n",
					 "    the leftmost column is column 1\n";
	}
	exit (1);
}

# --

while ($line = <STDIN>) {
	
	# Always pass the header lines through.
	if ($line eq $Header1 || $line eq $Header2) {
		print $line;
		next;
	}
	
	@tests = @theTests;
	
	$ok = 1;
	while ($ok && @tests > 0) {
		$test = shift (@tests);
		$arg = shift (@tests);
		
		@columns = split (/\t/, $line);
		
		if      ($test eq "cb") {
			$ok = 0 if ($arg <  0 || &MachineDate ($columns [4]) >  $arg);
		} elsif ($test eq "ca") {
			$ok = 0 if ($arg >= 0 && &MachineDate ($columns [4]) <= $arg);
		} elsif ($test eq "ub") {
			$ok = 0 if ($arg <  0 || &MachineDate ($columns [4]) >  $arg);
		} elsif ($test eq "ua") {
			$ok = 0 if ($arg >= 0 && &MachineDate ($columns [4]) <= $arg);
		} elsif ($test < 0) {
			$ok = 0 if ($columns [-$test-1] =~ /$arg/i);
		} else {
			$ok = 0 if ($columns [$test-1] !~ /$arg/i);
		}
	}
	print $line,"\n" if ($ok);
}

exit (0);

# --

# Process the command-line arguments, setting @theTests.
# Return 1 if all is well, 0 if there is any error in the command line.

sub ProcessArgs {
	return 0 if (@ARGV < 1);			# at least one argument required
	
	@theTests = ();
	
	local (@args) = ();
	local ($arg);
	
	# First, parse @ARGV into @args.
	local ($nextIsDate) = 0;			# whether next item may be part of a date
	
	while ($arg = shift (@ARGV)) {
	
		if ($arg =~ /\-([uc][ab])/) {
			push (@args, $1, "");
			$nextIsDate = 1;
			
		} elsif ($arg =~ /[\+\-][0-9]+/) {
			return 0 if (@ARGV < 1);
			push (@args, 0+$arg, shift (@ARGV));
			$nextIsDate = 0;
			
		} elsif ($nextIsDate) {
			$args [$#args] .= $arg." ";
			
		} else {
			return 0;
		}
	}
	
	# Now copy @args to @theTests, translating the dates into numeric times.
	while ($arg = shift (@args)) {
		push (@theTests, $arg);
		if ($arg =~ /[uc][ab]/) {
			local ($dateString) = shift (@args);
			local ($date) = &HumanDate ($dateString);
			if ($date == -1) {
				if ($dateString eq "") {
					$errorMessage = "missing date";
				} else {
					$errorMessage = "unrecognizable date '$dateString'";
				}
				return 0;
			}
			push (@theTests, $date);
		}
	}
	
	return 1;
}

# --

# HumanDate ($date)
# Given a human-written date string, convert it to a numeric system time.
# Return -1 if it isn't a valid date.

sub HumanDate {
	local ($date) = @_;
	
	return -1 if ($date eq "");
	
	local (@currentTime) = localtime (time);
	local ($month) = $currentTime [4];			# default month is this month
	local ($year) = $currentTime [5];			# default year is this year
	local ($temp);
	
	if ($date =~ /^\s*([0-9]+)[\W]*(([a-z]+)[\W]*([0-9]*))?\s*$/i) {		# 10 [Feb [1995]]
		$day = $1;
		$temp = $Months {substr ($3, 0, 3)};
		$month = $temp-1 if ($temp ne "");
		$year = $4 if ($4 ne "");
	} elsif ($date =~ /^\s*([a-z]+)[\W]*([0-9]+)[\W]*([0-9]*)\s*$/i) {		# Feb 10 [1995]
		$temp = $Months {substr ($1, 0, 3)};
		return -1 if ($temp eq "");
		$month = $temp - 1;
		$day = $2;
		$year = $3 if ($3 ne "");
	} elsif ($date =~ /^\s*([0-9]+)[\W]+([0-9]+)([\W]+([0-9]+))?\s*$/) {	# 2.10[.95]
		$month = $1 - 1;
		$day = $2;
		$year = $4 if ($4 ne "");
	} else {
		return -1;
	}
	
	return -1 if ($day < 1 || $day > 31);
	return -1 if ($month < 0 || $month > 11);
	
	$year -= 1900 if ($year > 1900);										# years since 1900
	
	return &timelocal (0, 0, 0, $day, $month, $year);
}

# --

# MachineDate ($date)
# Given a machine-written date string, as created by &ShortTimeString, convert it
# to a numeric system time. A date string looks like "Apr 7 12:20:02 1995".
# Return -1 if it isn't a valid date.

sub MachineDate {
	local ($date) = @_;
	
	local ($month, $day, $year) = ($date =~ /^([A-Z][a-z][a-z]) ([\d]+) \d\d\:\d\d\:\d\d \d\d(\d\d)$/);
	return -1 if ($month eq "");
	return -1 if ($day < 1 || $day > 31);
	
	$month = $Months {substr ($month, 0, 3)};
	return -1 if ($month eq "");
	$month -= 1;
	
	return &timelocal (0, 0, 0, $day, $month, $year);
}
