# "threads.pl" in the Forum News Gateway
# copyright 1994,1995 by Swarthmore College

# This file contains routines to maintain a database of threading info for
# a group. The information is separated into %ThreadDB{} and %ArticleDB{}.

# (In all places, message-ids do not have the angle brackets on them.)
# These items are per-article. They're heavily used, and need to be kept in memory.
# $ThreadDB{"1356"} = "[msgid of article 1356]";
# $ThreadDB{"[msgid of article 1356]","artnum"} = "1356";
# $ThreadDB{"[msgid of dummy article]","artnum"} = "dummy";
# $ThreadDB{"[msgid of article 1356]","parent"} = "[msgid of parent article of article 1356]";
# $ThreadDB{"[msgid of article 1356]","children"} = "[msgid of child 1 of article 1356] [msgid of child 2 of article 1356] etc... ";

# These are per-subject, used during threading. Should probably be in memory.
# $ThreadDB{"threadroots",$Subject) = "[msgid of root1]\t[msgid of root2]\t...\t";
# $ThreadDB{"threadsubject",$Subject) = "[user-presentable subject of thread (given canonical subject)]";

# These items are unique and small, so it doesn't matter where they are.
# $ThreadDB{"DBRange","low"} = "1300"; (Low article number in database)
# $ThreadDB{"DBRange","high"} = "1350"; (High article number in database)
# $ThreadDB{""} = "subject of thread1\tsubject of thread2\tsubject of thread3\t";
#		(Where threads are defined by common subjects in addition to References relationships)
#		(Subjects here are in a canonicalized form - all lower case, with Re:'s removed)
# $ThreadDB{"recheck"} = time() when the articles were last rechecked, if any

# These items are per-article and can be kept on disk.
# $ArticleDB{"1356","subject"} = "subject of article 1356";
# $ArticleDB{"1356","authorreal"} = "real name of author of article 1356";
# $ArticleDB{"1356","authoremail"} = "e-mail address of author of article 1356";
# $ArticleDB{"1356","date"} = "date of article 1356";
# $ArticleDB{"1356","numlines"} = "length in lines of article 1356";
#		(The range in the database may not be quite as up-to-date as the range on the server
#		 because we don't fetch all the new headers until we have to.)

# "threadroots" lists message IDs of articles which are the roots of disjoint
# subtrees in the same thread. "threadsubject" is the user-readable subject of the
# thread. The thread subject is updated when new articles arrive (and takes into account
# that articles have expired; if articles with a new subject appear in a thread, when
# all articles with the original articles have expired the thread subject will change).

# Note that in "children", "threadroots" and "" every item is followed by
# a delimiter, even the last item.
# So $ThreadDB{"MsgID","children"} might be "[child 1] [child 2] ", ending with a space, and
# $ThreadDB{""} might be "subject 1\tsubject 2\t", ending with a tab.

# Dummy articles:
# Dummy articles have message IDs but do not have article numbers. If you ask for the
# article number of a dummy article, $ThreadDB{[msgID],"artnum"}, you get "dummy".
# Since there's no article number, you can't do $ThreadDB{[article number]} to get the
# message ID. A dummy article may have a parent and children, but it does not have
# any other attributes: no subject, no author, no date, etc.
# Dummy articles substitute for real articles which no longer exist on the server, which
# have not yet been read from the server, or which never existed but were expected to because
# of a bogus or corrupted message ID.

# UpdateThreadDB()
# Gets headers for new articles from the server.  It puts the data in the
# current threadDB file (so you have to set the current group and get the low and
# high article numbers before you can call this routine).  It also gets rid of articles
# that have expired.

sub UpdateThreadDB {
	local($LowArticleNumber) = $CurrentGroupLowWaterMark;
	local($HighArticleNumber) = $CurrentGroupHighWaterMark;
	local($OldLowArticleNumber) = $ThreadDB{"DBRange","low"};
	local($OldHighArticleNumber) = $ThreadDB{"DBRange","high"};
	
#	print STDERR "Threading\n";
#	print STDERR "Old article range $OldLowArticleNumber-$OldHighArticleNumber\n";
#	print STDERR "New article range $LowArticleNumber-$HighArticleNumber\n";
	
	# Turn expired articles into dummies.
	if ($OldHighArticleNumber > 0) {
		local($I);
		if ($HighArticleNumber > 0) {
#			print STDERR "Dummifying $OldLowArticleNumber to ",$LowArticleNumber-1,"\n";
			for ($I = $OldLowArticleNumber; $I < $LowArticleNumber; $I++) {
				&DummifyArticle($I);
			}
		} else {
			# There's nothing left on the server; it's all expired.  Zap it all.
#			print STDERR "Dummifying all: $OldLowArticleNumber to $OldHighArticleNumber\n";
			for ($I = $OldLowArticleNumber; $I <= $OldHighArticleNumber; $I++) {
				&DummifyArticle($I);
			}
		}
	} else {
		# There wasn't anything in the DB before this update, so we don't need to dummify.
	}
	
	# Double-check the existence of articles now,
	# after trashing the old ones and before adding new ones,
	# so there's a little to do as possible.
	# Only do it if new articles have been added since the last time articles were threaded,
	# and if at least $RecheckInterval seconds have passed since the last recheck.
	local($Now) = time;
	if ($HighArticleNumber > $OldHighArticleNumber &&
			($ThreadDB{"recheck"} eq "" || $Now > $ThreadDB{"recheck"} + $RecheckInterval)) {
#	if (1) {
		$ThreadDB{"recheck"} = $Now;
		&RecheckArticles($LowArticleNumber,$OldHighArticleNumber);		# check between new low and old high
#		print STDERR "rechecking took ",time-$Now," seconds\n";
	} else {
#		print STDERR "not rechecking\n";
	}
	
	# Fetch new overview data, if necessary.
	local($NumNewArticles) = 0;
	if ($HighArticleNumber > $OldHighArticleNumber) {
		
		# Fetch everything we need, but nothing which does not exist.
		local($Low) = $OldHighArticleNumber+1;
		$Low = $LowArticleNumber if ($Low < $LowArticleNumber);
		
#		print STDERR "Fetching new articles $Low-$HighArticleNumber\n";
		
		local($UsingXOVER) = $UseXOVER && &SendXOVER($Low,$HighArticleNumber);
		$NumNewArticles = &ThreadArticles($UsingXOVER,$Low,$HighArticleNumber);
	}
	
	if ($NumNewArticles > 0) {
		# This routine must be called whenever articles in new threads arrive, to set
		# the thread subjects. It also cleans up after dummified articles; we don't
		# always call it when there are dummies, which saves time and doesn't hurt noticeably.
		&ThreadChores();
	}
	
	if ($HighArticleNumber > $OldHighArticleNumber) {
		$ThreadDB{"DBRange","high"} = $HighArticleNumber;
	}
	if ($LowArticleNumber > 0) {
		$ThreadDB{"DBRange","low"} = $LowArticleNumber;
	} else {
		$ThreadDB{"DBRange","low"} = $ThreadDB{"DBRange","high"}+1;
	}
	
#	&DebugPrintArticleTree();
	
	1;
}


# DummifyArticle($ArticleNum)
# Turn the given article into a dummy. It has expired on the server.
# Dummy articles have article ID "dummy". They retain their parents and children,
# but lose their other attributes.

sub DummifyArticle {
	local($ArticleNum) = @_;
	local($MessageID) = $ThreadDB{$ArticleNum};
	
	delete $ThreadDB{$ArticleNum};
	$ThreadDB{$MessageID,"artnum"} = "dummy";
	delete $ArticleDB{$ArticleNum,"subject"};
	delete $ArticleDB{$ArticleNum,"authorreal"};
	delete $ArticleDB{$ArticleNum,"authoremail"};
	delete $ArticleDB{$ArticleNum,"numlines"};
	delete $ArticleDB{$ArticleNum,"date"};
}

# ThreadChores()
# Do two chores:
# (1) delete threads which consist entirely of dummy articles
# (2) update thread root subjects, $ThreadDB{"threadsubject",$CanonicalThreadSubject}.
# Since this is the only place that thread root subjects are set, it must be
# called whenever articles in new threads arrive. 

sub ThreadChores {
	local(@Subjects) = split(/\t/,$ThreadDB{""});		# canonical subjects
	local($Subject);
	local($ThreadSubject);								# user-visible thread subjects
	
	foreach $Subject (@Subjects) {
		$ThreadSubject = &NonemptyThread($Subject);
		if ($ThreadSubject eq "") {
			&DeleteThread($Subject);
		} else {
			$ThreadDB{"threadsubject",$Subject} = $ThreadSubject;
		}
	}
}

# NonemptyThread($Subject)
# If the given thread contains a non-dummy article, return the subject
# of the first non-dummy article found (or "[no subject]" if the subject is "");
# otherwise return "".

sub NonemptyThread {
	local($Subject) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Result) = "";
	
	foreach $Root (@Roots) {
		$Result = &NonemptySubthread($Root);
		last if ($Result ne "");
	}
	
	$Result;
}

# NonemptySubthread($Article)
# If the given subthread headed contains a non-dummy article, return
# the subject of the first non-dummy article found (or "[no subject]" if the subject
# is ""); otherwise return "".

sub NonemptySubthread {
	local($ArticleID) = @_;
	local($ArticleNum) = $ThreadDB{$ArticleID,"artnum"};
	local($Result);
	
	if ($ArticleNum ne "dummy") {
		$Result = $ArticleDB{$ArticleNum,"subject"};
		$Result = "[no subject]" if ($Result eq "");
		return $Result;
	}
	
	local(@Children) = split(/ /,$ThreadDB{$ArticleID,"children"});
	local($Child);
	foreach $Child (@Children) {
		$Result = &NonemptySubthread($Child);
		return $Result if ($Result ne "");
	}
	
	"";												# no non-dummy articles
}

# DeleteThread($Subject)
# Delete the given thread from %ThreadDB. All the articles are dummies.

sub DeleteThread {
	local($Subject) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Root);
	
	foreach $Root (@Roots) {
		&DeleteSubthread($Root);
	}
	
	delete $ThreadDB{"threadroots",$Subject};
	delete $ThreadDB{"threadsubject",$Subject};
	$Subject =~ s/(\W)/\\$1/g; # put backslashes in front of funky chars so we can use this as a pattern
	$ThreadDB{""} =~ s/(^|\t)$Subject\t/$1/;
}

# DeleteSubthread($Article)
# Delete the given subthread from %ThreadDB. All the articles are dummies.

sub DeleteSubthread {
	local($Article) = @_;
	
	local(@Children) = split(/ /,$ThreadDB{$Article,"children"});
	local($Child);
	foreach $Child (@Children) {
		&DeleteSubthread($Child);
	}
	
	delete $ThreadDB{$Article,"artnum"};
	delete $ThreadDB{$Article,"parent"};
	delete $ThreadDB{$Article,"children"};
}

# ValidReferences($ReferenceString)
# Given a string of references from an article, return a
# list of the apparently valid references in reverse order.
# This is used in threading articles. Also, the references we display for an
# article are the ones the article supplies, so we use it there too.

sub ValidReferences {
	local($ReferenceString) = @_;
	local(@References) = ();
	
	# References may be invalid. We have to check them more carefully than other fields.
	# Some references may have a missing "<" or ">" or be delimited by "," instead of " ".
	# We try to tolerate this kind of minor disturbance.
	# References that have other problems are ruthlessly ignored.
	# Note that we're reversing the list as we walk through it.
	local(@DubiousReferences) = split(/[\,\s]+/,$ReferenceString);
	local($Reference);
	while (@DubiousReferences > 0) {
		$Reference = shift(@DubiousReferences);
		$Reference =~ s/^<//;
		$Reference =~ s/>$// || $Reference =~ s/\.$//;		# final . is probably a typo for >
		if ($Reference =~ /^[^<>,\@]+\@[^<>,\@]+$/) {
			unshift(@References,$Reference);
		} else {
			# Invalid references are so common that they're not worth printing a message for.
			# print STDERR "**** Message $MessageID, invalid reference $Reference\n";
		}
	}
	@References;
}

# ThreadArticles($UsingXOVER,$LowArticle,$HighArticle)
# Fetch the header information for articles in the range $LowArticle to $HighArticle,
# and thread them. Return the number of new articles found.
# If $UsingXOVER is 1, then read the headers with XOVER.
# If $UsingXOVER is 0, then read the headers with &GetFirstArticleOverview and &GetNextArticleOverview.
# They return the headers in different formats, so there's some translation work.

sub ThreadArticles {
	local($UsingXOVER,$LowArticle,$HighArticle) = @_;
	local($Header);				# header line for XOVER
	local(@Headers);			# headers list for the Other Way
	local($NumArticles) = 0;
	
	# Fetch the first article's headers.
	if ($UsingXOVER) {
		$Header = <NNTPSERVER>;
		return if ($Header eq ".\015\012");				# no work to do
		&ThreadXOVERArticle($Header);
		++$NumArticles;
	} else {
		@Headers = &GetFirstArticleOverview($LowArticle,$HighArticle);
		return if (@Headers == 0);						# no work to do
		&ThreadOtherWayArticle(@Headers);
		++$NumArticles;
	}
	
	while(1) {
		if ($UsingXOVER) {
			$Header = <NNTPSERVER>;
			last if ($Header eq ".\015\012");				# finished
			&ThreadXOVERArticle($Header);
			++$NumArticles;
		} else {
			@Headers = &GetNextArticleOverview($HighArticle);
			last if (@Headers == 0);						# finished
			&ThreadOtherWayArticle(@Headers);
			++$NumArticles;
		}
	}
	
	return $NumArticles;
}

# ThreadXOVERArticle($Header)
# $Header is an article header line as returned by XOVER.
# Thread the article.

sub ThreadXOVERArticle {
	local($Header) = @_;
	
	if ($Header !~ /([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t\015\012]*)/) {
		print STDERR "In ThreadXOVERArticle: Mangled overview line: $Header\n";
		return;
	}
	
	&ThreadArticle($1,$2,$3,$4,$5,$6,$8);
}

# ThreadOtherWayArticle(@Headers)
# @Headers is a list of the headers of an article as returned by &GetFirstArticleOverview
# or &GetNextArticleOverview (the Other Way of getting article headers).
# Thread the article.

sub ThreadOtherWayArticle {
	local(@Headers) = @_;
	local($HeaderLine);
	local($ArticleNum,$Subject,$From,$Date,$MessageID,$References,$LineLen);
	
	$ArticleNum = "";
	$Subject = "";
	$From = "";
	$Date = "";
	$MessageID = "";
	$References = "";
	$LineLen = "";
	
#	print STDERR "-\n";
	for $HeaderLine (@Headers) {
#		print STDERR "|$HeaderLine";
		# The Article-Num: field is inserted by the routine that fetches the headers.
		$HeaderLine =~ /^Article-Num:\s*([0-9]+)/      && ($ArticleNum = $1, next);
		$HeaderLine =~ /^Subject:\s*([^\015\012]+)/    && ($Subject = $1,    next);
		$HeaderLine =~ /^From:\s*([^\015\012]+)/       && ($From = $1,       next);
		$HeaderLine =~ /^Date:\s*([^\015\012]+)/       && ($Date = $1,       next);
		$HeaderLine =~ /^Message-ID:\s*([^\015\012]+)/ && ($MessageID = $1,  next);
		$HeaderLine =~ /^References:\s*([^\015\012]+)/ && ($References = $1, next);
		$HeaderLine =~ /^Lines:\s*([0-9]+)/            && ($LineLen = $1,    next);
	}
	
	&ThreadArticle($ArticleNum,$Subject,$From,$Date,$MessageID,$References,$LineLen);
}

# ThreadArticle($ArticleNum,$Subject,$From,$Date,$MessageID,$References,$LineLen)
# This routine takes the given article and inserts it into %ThreadDB.

sub ThreadArticle {
	local($ArticleNum,$Subject,$From,$Date,$MessageID,$References,$LineLen) = @_;
	
#	print STDERR "-\n";
#	print STDERR "ArticleNum $ArticleNum\n";
#	print STDERR "Subject $Subject\n";
#	print STDERR "From $From\n";
#	print STDERR "Date $Date\n";
#	print STDERR "MessageID $MessageID\n";
#	print STDERR "References $References\n";
#	print STDERR "LineLen $LineLen\n";
	
	# Remove <angle brackets> from the message ID. */
	if ($MessageID =~ /^<([^<>,\@\s]+\@[^<>,\@\s]+)>$/) {
		$MessageID = $1;
	} else {
		print STDERR "Invalid message ID $MessageID in $CurrentGroup; article skipped.\n";
		return;
	}
	
	local($HasChildren) = 0;				# new non-dummy articles do not have children
	
	# See if this msgid is already in the threadDB as a dummy article.
	local($ArticleNumInDB) = $ThreadDB{$MessageID,"artnum"};
	if ($ArticleNumInDB eq "dummy") {
		# It is. Unlink the dummy from its parent, on the assumption that we should
		# trust the real article's references, rather than the article that created the dummy.
		&DeleteChildPointer($MessageID);
		$ThreadDB{$MessageID,"parent"} = "";
		
		# This is the only case in which a new article may already have children.
		$HasChildren = $ThreadDB{$MessageID,"children"} ne "";
	} elsif ($ArticleNumInDB ne "") {
		print STDERR "In ThreadArticle: article to be inserted already existed as a non-dummy article.\n";
#		print STDERR "Article number $ThreadDB{$MessageID,\"artnum\"} <$MessageID>\n";
		return;
	}
	
	$ThreadDB{$ArticleNum} = $MessageID;
	$ThreadDB{$MessageID,"artnum"} = $ArticleNum;
	$ArticleDB{$ArticleNum,"subject"} = $Subject;
	$ArticleDB{$ArticleNum,"date"} = $Date;
	$ArticleDB{$ArticleNum,"numlines"} = $LineLen;
	
	# Figure out real name and e-mail address of author.
	if ($From =~ /^(\S+)\s+\(([^\)]*)\)/) {
		$ArticleDB{$ArticleNum,"authoremail"} = $1;
		$ArticleDB{$ArticleNum,"authorreal"} = $2;
		
		# Correct broken names and e-mail addresses which include < >.
		$ArticleDB{$ArticleNum,"authoremail"} =~ tr/\<\>//d;
		$ArticleDB{$ArticleNum,"authorreal"} =~ tr/\<\>//d;
		
	} elsif ($From =~ /^([^<]*)<([^>]*)>/) {
		$ArticleDB{$ArticleNum,"authoremail"} = $2;
		# Remove trailing whitespace from the real name.
		local($RealName) = $1;
		$RealName = reverse($RealName);
		$RealName =~ s/^\s+//;
		$ArticleDB{$ArticleNum,"authorreal"} = reverse($RealName);
	} else {
		# In this case, we do not set "authorreal".
		# Correct broken e-mail addresses which include < >.
		$From =~ tr/\<\>//d;
		$ArticleDB{$ArticleNum,"authoremail"} = $From;
	}
	
	# Go through the list of references, linking up parents and children.
	# Create dummy articles to stand for novel message IDs.
	# Skip duplicate references, which would otherwise cause loops in the article tree.
	# References to the original article are considered "duplicates".
	local(@References) = &ValidReferences($References);
	local($TopOfNewChain) = $MessageID;
	local($Reference,$TheReference);		# $TheReference is to hold $Reference after the loop ends
	local($ReferenceExisted) = 0;
	local(%Refs) = ($MessageID,"x");
	foreach $Reference (@References) {
		if ($ThreadDB{$Reference,"artnum"} ne "") {
			# Ignore duplicate references. Duplicate references are rare, and if we have one
			# then we will reach this fork of the if statement, so it's more efficient to
			# put the test inside here than to leave it at the top level of the loop.
			next if ($Refs{$Reference} ne "");
			
			# The top of the new chain so far is going to be the child of the reference.
			# Don't insert the child yet, because there's a chance it may cause a loop in
			# the article tree. For now, just remember that the reference existed, and
			# retain the $Reference and $TopOfNewChain variables. Below we check to see
			# whether there was a loop, and if not, we insert the chain into the tree.
			$TheReference = $Reference;		# $Reference goes away after the loop exits, but $TheReference stays
			$ReferenceExisted = 1;
			last;
		} else {
			# Create dummy article for the reference with top of new chain so far as its child.
			$ThreadDB{$Reference,"artnum"} = "dummy";
			&InsertChildPointer($Reference,$TopOfNewChain);
			$ThreadDB{$TopOfNewChain,"parent"} = $Reference;
			$TopOfNewChain = $Reference;
			$Refs{$Reference} = "x";			# remember this reference, in case it is duplicated later
		}
	}
	
	if ($ReferenceExisted) {
		# One of the references was an existing article.
		# If the article being threaded has children, then it is possible that the last of
		# the article's references was to one of its children. That would cause a loop in
		# the article tree. The reference loop stops when it reaches any existing article, so
		# no other reference could cause a loop, and there can be at most one loop.
		
		# Variable $TheReference is set only if $ReferenceExisted is true, so it should be
		# used only inside this if statement.
		
		if ($HasChildren) {					# a loop is possible only if the article has children
			local($ID) = $TheReference;		# message ID of reference
			local($Parent);
			for (;;) {
				$Parent = $ThreadDB{$ID,"parent"};
				last if ($Parent eq "");	# no loop in the article tree
				if ($Parent eq $MessageID) {
					# The reference existed, but it caused a loop so we're going to pretend it didn't.
					# Below, we'll create a new thread root to hold this thread. For that, retain
					# the value of $TopOfNewChain.
					print STDERR "Found loop to child from $TopOfNewChain to $TheReference\n";
					$ReferenceExisted = 0;
					last;
				}
				$ID = $Parent;
			}
		}
		
		if ($ReferenceExisted) {
			# There is no loop. Insert the chain into the article tree.
			# $TheReference is an existing article in the tree, and $TopOfNewChain
			# is the top end of the chain of dummied-up references, or if there
			# are no dummy references, it's the article we're threading now.
			&InsertChildPointer($TheReference,$TopOfNewChain);
			$ThreadDB{$TopOfNewChain,"parent"} = $TheReference;
		}
	}
	
	if (!$ReferenceExisted) {
		# We went through all the references and did not find an existing article.
		# That means we didn't create a loop in the tree of articles.
		# If there are no references, we end up here.
		# Make the new article chain a thread root.
		&InsertChildPointer("",$TopOfNewChain);
		$ThreadDB{$TopOfNewChain,"parent"} = "";
	}
	
#	print STDERR "Done threading this article.\n";
}

# CanonicalizeSubject($Subject)
# Returns subject lowercased, with leading Re: removed.
# Characters which are not allowed in URLs are replaced with '_',
# and '+' is also replaced with '_' so that it won't turn into a space.
# These canonical subjects appear in /Newsreading/ThreadPages/... URLs.

sub CanonicalizeSubject {
	local($Subject) = @_;
	
	if ($Subject eq "") {
		"[no subject]";
	} else {
		$Subject =~ tr/A-Z/a-z/;
		$Subject =~ s/^re:\s+//;
#		$Subject =~ tr/<>%"=;\/#?: +/_/;
		$Subject =~ tr/a-zA-Z0-9\$\-\.\&\!\*\'\(\)\,/_/c;
		$Subject;
	}
}

# InsertChildPointer($ParentMsgID,$ChildMsgID)
# Make the $ChildMsgID article a child of the $ParentMsgID article.
# If $ParentMsgID is "", then $ChildMsgID is to be made a thread root,
# that is, a top-level article in whatever thread it belongs to.
# If it does not belong to an existing thread, create a thread for it to
# belong to.
# This does not keep children sorted by date. Maybe we could do that someday.

sub InsertChildPointer {
	local($ParentMsgID,$ChildMsgID) = @_;
	
	if ($ParentMsgID ne "") {
		# There is a parent. Add a new child.
		local($ChildMsgIDPattern) = $ChildMsgID;
		$ChildMsgIDPattern =~ s/(\W)/\\$1/g; # put backslashes in front of funky chars so we can use this as a pattern
		
		# If child is not already there, tack it on at the end.
		if ($ThreadDB{$ParentMsgID,"children"} !~ /$ChildMsgIDPattern /) {		# the space is necessary!
			$ThreadDB{$ParentMsgID,"children"} .= $ChildMsgID." ";
		}
		
	} else {
		# There is no parent. Create a new root, or add it to an existing one.
		local($NewSubject) = &DoSubjectOfThread($ChildMsgID);
		$NewSubject = &CanonicalizeSubject($NewSubject);
		
		local($NewSubjectPattern) = $NewSubject;
		$NewSubjectPattern =~ s/(\W)/\\$1/g; # put backslashes in front of funky chars so we can use this as a pattern
		
		if ($ThreadDB{""} =~ /$NewSubjectPattern\t/) {
			# The subject is already at the top level. Add a new root to the thread's list of roots.
			$ThreadDB{"threadroots",$NewSubject} .= $ChildMsgID."\t";
		} else {
			# The subject is not at the top level. Add it in there, and start a list of roots for the new thread.
			$ThreadDB{""} .= $NewSubject."\t";
			$ThreadDB{"threadroots",$NewSubject} = $ChildMsgID."\t";
		}
	}
}

# DeleteChildPointer($ChildMsgID)
# Remove this child article from its parent's knowledge.
# This has the effect of unlinking the child and everything below it
# from the article tree in the ThreadDB.

sub DeleteChildPointer {
	local($ChildMsgID) = @_;
	
	$ParentMsgID = $ThreadDB{$ChildMsgID,"parent"};
	
	if ($ParentMsgID ne "") {
		# The child has a parent. Remove the child from the parent's list of children.
		
		$ChildMsgID =~ s/(\W)/\\$1/g; # put backslashes in front of funky chars so we can use this as a pattern
#		$ThreadDB{$ParentMsgID,"children"} =~ s/(^| )$ChildMsgID /$1/;
		$ThreadDB{$ParentMsgID,"children"} =~ s/(^| )$ChildMsgID /$1/g == 1
			|| print STDERR "DeleteChildPointer, wrong # kids\n";
		
	} else {
		# The child does not have a parent. It is a thread root.
		# Remove the child from the list of thread roots that it appears on.
		
		local($Subject) = &DoSubjectOfThread($ChildMsgID);
		$Subject = &CanonicalizeSubject($Subject);
		
		if ($ThreadDB{"threadroots",$Subject} eq $ChildMsgID."\t") {
			# This root is the only one in the thread. Delete the thread.
			delete $ThreadDB{"threadroots",$Subject};
			
			# Delete the subject from ThreadDB.
			$Subject =~ s/(\W)/\\$1/g; # put backslashes in front of funky chars so we can use this as a pattern
#			$ThreadDB{""} =~ s/(^|\t)$Subject\t/$1/;
#			$ThreadDB{""} =~ s/(^|\t)$Subject\t/$1/g == 1 || print STDERR "DeleteChildPointer, wrong # threads $Subject.\n";
			local($NThreads) = $ThreadDB{""} =~ s/(^|\t)$Subject\t/$1/g;
			if ($NThreads != 1) {
				print STDERR "DeleteChildPointer, $NThreads thread roots in subject \"$Subject\".\n";
				print STDERR "Group \"$CurrentGroup\" on \"$CurrentServer\".\n"
			}
		} else {
			# This root is not the only one in the thread. Find it and remove it.
			local($ChildPattern) = $ChildMsgID;
			$ChildPattern =~ s/(\W)/\\$1/g; # put backslashes in front of funky chars so we can use this as a pattern
			
#			$ThreadDB{"threadroots",$Subject} =~ s/(^|\t)$ChildPattern\t/$1/;
			$ThreadDB{"threadroots",$Subject} =~ s/(^|\t)$ChildPattern\t/$1/g == 1
				|| print STDERR "DeleteChildPointer, wrong # roots\n";
		}
	}
}

# ------------------------------------------------------------------------------

# RecheckArticles($LowArticleNum,$HighArticleNum)
# Check which articles in the current group (still) exist on the news server.
# It's only necessary to check between $LowArticleNum and $HighArticleNum.
# Dummify articles that have gone missing.

sub RecheckArticles {
	local($LowArticleNum,$HighArticleNum) = @_;
	
	# Do nothing if there's nothing to do.
	# This can happen when we're called before any articles have been threaded.
	return if ($LowArticleNum > $HighArticleNum);
	
#	&TestRecheck($LowArticleNum,$HighArticleNum) && return;
	
	local($UsingXOVER) = $UseXOVER && &SendXOVER($LowArticleNum,$HighArticleNum);
	if ($UsingXOVER) {
		&XOVERRecheck();
	} else {
		&SlowRecheck();
	}
}

# TestRecheck()
# Recheck all articles the ultracautious way, reporting anything unexpected.

sub TestRecheck {
	local($Low,$High) = @_;
	local($ArticleNum);
	local($ResponseCode,$ResponseText);
	local($Exists);
	local($MessageID);
	
	print STDERR "test rechecking from $Low - $High\n";
	
	for ($ArticleNum = $Low; $ArticleNum <= $High; ++$ArticleNum) {
		($ResponseCode,$ResponseText) = &SendNNTPCommand("STAT $ArticleNum\012");
		$Exists = ($ResponseCode == 223);		# whether the article exists
		$MessageID = $ThreadDB{$ArticleNum};
		if ($Exists) {
			if ($MessageID eq "") {
				# The article exists and we don't know it.
				print STDERR "-- unexpected article $ArticleNum\n";
			}
		} else {
			if ($MessageID ne "") {
				# The article does not exist, but we have a record of it.
				# print STDERR "** (test) missing article $ArticleNum <$MessageID>\n";
				&DummifyArticle($ArticleNum);
			}
		}
	}
}

# XOVERRecheck()
# Use XOVER to recheck articles more quickly.

sub XOVERRecheck {
	local(%OKArticles) = ();
	local($Header);
	
	# Do this calculation first. Hopefully it will overlap the NNTP server's
	# response time to the XOVER that was sent out by &RecheckArticles, above,
	# shaving PRECIOUS MICROSECONDS from our own response time. :-/
	local(@Subjects) = split(/\t/,$ThreadDB{""});		# canonical subjects
	local($Subject);
	local(@Roots);
	
	$OKArticles{"dummy"} = "x";							# dummy articles are already OK
	while (1) {
		$Header = <NNTPSERVER>;
		last if ($Header eq ".\015\012");
		if ($Header !~ /([^\t]*)\t[^\t]*\t[^\t]*\t[^\t]*\t[^\t]*\t[^\t]*\t[^\t]*\t[^\t\015\012]*/) {
			print STDERR "In XOVERRecheck: Mangled overview line: $Header\n";
			return;
		}
		$OKArticles{$1} = "x";
	}
	
	foreach $Subject (@Subjects) {
		local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
		foreach $Root (@Roots) {
			&XOVERRecheckSubthread($Root);
		}
	}
}

# XOVERRecheckSubthread($MessageID)
# Check the existence of all articles in the subthread.
# This inherits the variable %OKArticles from its caller, XOVERRecheck.

sub XOVERRecheckSubthread {
	local($MessageID) = @_;
	local($ArticleNum) = $ThreadDB{$MessageID,"artnum"};
	
	if ($OKArticles{$ArticleNum} eq "") {
		print STDERR "-> (XOVER) missing article $ArticleNum <$MessageID>\n";
		&DummifyArticle($ArticleNum);
	}
	
	local(@Children) = split(/ /,$ThreadDB{$MessageID,"children"});
	local($Child);
	foreach $Child (@Children) {
		&XOVERRecheckSubthread($Child);
	}
}

# SlowRecheck()
# Recheck all articles, the slow way.

sub SlowRecheck {
	local(@Subjects) = split(/\t/,$ThreadDB{""});		# canonical subjects
	local($Subject);
	local(@Roots);
	
	foreach $Subject (@Subjects) {
		local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
		foreach $Root (@Roots) {
			&SlowRecheckSubthread($Root);
		}
	}
}

# SlowRecheckSubthread($MessageID)
# Check the existence of all articles in the subthread.

sub SlowRecheckSubthread {
	local($MessageID) = @_;
	local($ArticleNum) = $ThreadDB{$MessageID,"artnum"};
	
	if ($ArticleNum ne "dummy" && !&DoesArticleExist($MessageID)) {
		print STDERR "-> (slow) missing article $ArticleNum <$MessageID>\n";
		&DummifyArticle($ArticleNum);
	}
	
	local(@Children) = split(/ /,$ThreadDB{$MessageID,"children"});
	local($Child);
	foreach $Child (@Children) {
		&SlowRecheckSubthread($Child);
	}
}

1;
